// This program reads the bytes from a .bmp file and manipulates the
// bytes to create a new bitmap.

#include <iostream>
#include <fstream>
#include <iomanip>
#include <time.h>
#include <string>
using namespace std;

// Represents the red, green, and blue components of a single pixel
class Pixel
{
public:
	Pixel(unsigned char red, unsigned char green, unsigned char blue);
	unsigned char blue;
	unsigned char green;
	unsigned char red;
};

Pixel::Pixel(unsigned char red = 0, unsigned char green = 0, unsigned char blue = 0)
{
	this->red = red;
	this->green = green;
	this->blue = blue;
}

/*********************************************************************************/

// Represents an entire .bmp image
class Bitmap
{
public:
	Bitmap() { width = 0; height = 0; }
	bool OpenFile(string filename);
	void SaveFile(string filename);
	void SetPixel(int row, int col, Pixel pixel);
	Pixel GetPixel(int row, int col);
	int GetWidth() { return width; }
	int GetHeight() { return height; }
	void Print();
	//void Blur();
	
private:
	int width;
	int height;
	Pixel* pixels;
	char headers[54];
};

bool Bitmap::OpenFile(string filename)
{
	ifstream fin;
	fin.open(filename, ios::binary);
	if (fin.fail())
	{
		cout << "Failed to open " << filename << endl;
		return false;
	}
		
	// Read first 54 bytes that form the file and DIB headers
	fin.read(headers, 54);

	// Read width and height from 18 byte-offset
	fin.seekg(18);
	fin.read((char*)&width, 4);
	fin.read((char*)&height, 4);

	// Allocate memory for bitmap's pixels
	pixels = new Pixel[width * height];

	// Move to location of pixels
	fin.seekg(54);  
	int i = 0;
	for (unsigned int row = 0; row < height; row++)
	{
		for (unsigned int column = 0; column < width; column++, i++)
			fin.read((char*)&(pixels[i]), 3);

		// Read remaining pad bytes
		char junk;
		for (unsigned int bytesRead = width * 3; bytesRead % 4 != 0; bytesRead++)
			fin.read(&junk, 1);			
	}

	fin.close();
	return true;
}

void Bitmap::SaveFile(string filename)
{
	ofstream fout;
	fout.open(filename, ios::binary);   // destroys the old file

	// Write file and DIB headers
	fout.write(headers, 54);

	// Move to location of pixels
	//fout.seekp(54);  
	int i = 0;
	for (unsigned int row = 0; row < height; row++)
	{
		for (unsigned int column = 0; column < width; column++, i++)
			fout.write((char*)&(pixels[i]), 3);

		// Write remaining pad bytes
		char junk = 0;
		for (int bytesWritten = width * 3; bytesWritten % 4 != 0; bytesWritten++)
			fout.write(&junk, 1);
	}

	fout.close();
}

// Sets pixel in bitmap at the given row and col where row 0 is the top row and
// col 0 is the left-most column.
void Bitmap::SetPixel(int row, int col, Pixel pixel)
{
	int pixelCount = width * height;
	if (row * width + col < pixelCount)
		pixels[(height - row - 1) * width + col] = pixel;
	else
		cout << "Warning: " << row << ", " << col << " are out of bounds in SetPixel\n";
}

// Gets pixel from bitmap at the given row and col where row 0 is the top row and
// col 0 is the left-most column.
Pixel Bitmap::GetPixel(int row, int col)
{
	int pixelCount = width * height;
	if (row * width + col < pixelCount)
		return pixels[(height - row - 1) * width + col];
	else
	{
		cout << "Warning: " << row << ", " << col << " are out of bounds in GetPixel\n";
		return pixels[0];
	}
}

void Bitmap::Print()
{
	cout << "Width = " << width << "\nHeight = " << height << endl;
}

/*********************************************************************************/

void InvertColors(Bitmap &bitmap)
{
	// Loop through all pixels and invert RGB. So if blue = 255 then new
	// blue = 0, if red = 100 then new red = 155, etc.
	for (int r = 0; r < bitmap.GetHeight(); r++)
	{
		for (int c = 0; c < bitmap.GetWidth(); c++)
		{
			Pixel pixel = bitmap.GetPixel(r, c);
			pixel.red = abs(pixel.red - 255);
			pixel.green = abs(pixel.green - 255);
			pixel.blue = abs(pixel.blue - 255);
			bitmap.SetPixel(r, c, pixel);
		}
	}
}

void BlackAndWhite(Bitmap &bitmap)
{
	for (int r = 0; r < bitmap.GetHeight(); r++)
	{
		for (int c = 0; c < bitmap.GetWidth(); c++)
		{
			Pixel pixel = bitmap.GetPixel(r, c);
			int ave = (pixel.red + pixel.green + pixel.blue) / 3;
			if (ave < 120)
			{
				pixel.red = 0;
				pixel.green = 0;
				pixel.blue = 0;
			}
			else
			{
				pixel.red = 255;
				pixel.green = 255;
				pixel.blue = 255;
			}

			bitmap.SetPixel(r, c, pixel);

		}
	}
}

int main()
{
	cout << "Filename? ";
	string filename;
	getline(cin, filename);

	Bitmap bitmap;
	Pixel pixel;

	if (bitmap.OpenFile(filename))
	{
		bitmap.Print();
		//Menu(bitmap);

		//InvertColors(bitmap);
		BlackAndWhite(bitmap);

		// Save transformed bitmap with -NEW.bmp name
		string newNilename = filename.substr(0, filename.find(".")) + "-NEW.bmp";
		cout << "\nWriting to " << newNilename << endl;
		bitmap.SaveFile(newNilename);
	}

	return 0;
}
